#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ply.yacc as yacc

from pyxie.parsing.lexer import tokens

class Grammar(object):
    tokens = tokens
    def p_error(self,p):
        print "Syntax error at", p

    def p_program(self, p):
        "program : statements"
        p[0] = [ "program", p[1] ]

    def p_statements_1(self, p):
        "statements : statement"
        p[0] = [ "statements", [ p[1][1] ]]

    def p_statements_2(self, p):
        "statements : statement statements"
        p[0] = [ "statements", [ p[1][1] ] + p[2][1] ]

    def p_statement_1(self, p):
        "statement : assignment_statement EOL"
        p[0] = [ "statement", p[1] ]

    def p_statement_2(self, p):
        "statement : print_statement EOL"
        p[0] = [ "statement", p[1] ]

    def p_print_statement_1(self, p):
        "print_statement : PRINT expr_list"
        p[0] = [ "print_statement", p[2][1] ]


    def p_expr_list_1(self,p):
        "expr_list : value_literal"
        p[0] = [ "expr_list", [ p[1] ] ]

    def p_expr_list_2(self,p):
        "expr_list : value_literal COMMA expr_list"
        p[0] = [ "expr_list", [ p[1] ] + p[3][1] ]

    def p_statement_3(self, p):
        "statement : value_literal EOL"
        p[0] = [ "statement", p[1] ]

    def p_assignment_statement(self, p):
        "assignment_statement : IDENTIFIER ASSIGN value_literal"
        p[0] = ["assignment_statement", [p[1], "IDENTIFIER", p.lineno(1) ],
                                        ["ASSIGN", p[2]],
                                        p[3]
               ]

    def p_value_literal_1(self, p):
        "value_literal : NUMBER"
        p[0] = [ "value_literal", p[1], "NUMBER", "INT", p.lineno(1) ]

    def p_value_literal_2(self, p):
        "value_literal : FLOAT"
        p[0] = [ "value_literal", p[1], "FLOAT", p.lineno(1) ]

    def p_value_literal_3(self, p):
        "value_literal : HEX"
        p[0] = [ "value_literal", p[1], "NUMBER", "HEX", p.lineno(1) ]

    def p_value_literal_4(self, p):
        "value_literal : OCTAL"
        p[0] = [ "value_literal", p[1], "NUMBER", "OCTAL", p.lineno(1) ]

    def p_value_literal_5(self, p):
        "value_literal : BINARY"
        p[0] = [ "value_literal", p[1], "NUMBER", "BINARY", p.lineno(1) ]

    def p_value_literal_6(self, p):
        "value_literal : STRING"
        p[0] = [ "value_literal", p[1], "STRING", p.lineno(1) ]

    def p_value_literal_7(self, p):
        "value_literal : BOOLEAN"
        p[0] = [ "value_literal", p[1], "BOOLEAN", "INT", p.lineno(1) ]

    def p_value_literal_8(self, p):
        "value_literal : IDENTIFIER"
        p[0] = [ "value_literal", p[1], "IDENTIFIER", p.lineno(1) ]

def parse(source,lexer):
   yacc.yacc(module=Grammar())
   result = yacc.parse(source, lexer=lexer)
   return result



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

def get_statements(AST):
    statements = []
    if AST[0] == "program":
        assert AST[1][0] == "statements"
        statements = AST[1][1]
    else:
        print AST[0]
    return statements

def todo(*args):
    print "TODO", " ".join([repr(x) for x in args])

def find_variables(AST):
    variables = {}
    statements = get_statements(AST)
    for statement in statements:
        tag, rest = statement[0], statement[1:]
        if tag == "assignment_statement":
            lvalue, assigntype, rvalue = rest
            if assigntype[1] != "=":
                todo("assignment where the assigntype is not '='")
                continue # Skip
            if lvalue[1] != "IDENTIFIER":
                todo("assignment where the lvalue is not an identifier")
                continue # Skip
            if rvalue[0] != "value_literal":
                todo("assignment where the rvalue is not a value_literal")
                continue # Skip

            identifer = lvalue[0]
            v_type = rvalue[2]
            if identifer in variables:
                todo("we could check that the identifier doesn't change type")
                continue # Skip
            variables[identifer] = v_type

    return variables

class UnknownType(Exception):
    pass

class CannotConvert(Exception):
    pass

def python_type_to_c_type(ptype):
    if ptype == "STRING":  return "string"
    if ptype == "NUMBER":  return "int"
    if ptype == "BOOLEAN": return "bool"
    if ptype == "FLOAT":   return "double"
    raise UnknownType("Cannot identify C Type for %s" % ptype)

def includes_for_ctype(ctype):
    if ctype == "string":  return "<string>"

def includes_for_cstatement(cstatement):
    if cstatement[0] == "print_statement": return "<iostream>"


def convert_assignment(assignment):
    lvalue, assigntype, rvalue = assignment

    if assigntype[1] != "=":
        todo("Convert Assignment where assigntype is not '='")
        raise CannotConvert("Cannot convert assignment where assigntype is not '='")
    if lvalue[1] != "IDENTIFIER":
        todo("assignment where the lvalue is not an identifier")
        raise CannotConvert("Cannot convert assignment where the lvalue is not an identifier")
    if rvalue[0] != "value_literal":
        todo("assignment where the rvalue is not a value_literal")
        raise CannotConvert("Cannot convert assignment where the rvalue is not a value_literal")

    lvalue = lvalue[0]
    rvalue = rvalue[1]
    return ["assignment", lvalue, "=", '"'+rvalue+'"' ]

def convert_print(arg_spec):
    arg_spec = arg_spec[0]
    cstatement = []
    cargs = []
    print "arg_spec",arg_spec[0]
    for arg in arg_spec:
        print arg[0]
        if arg[0] != "value_literal":
            todo("Handle print for non-value-literals")
            raise CannotConvert("Cannot convert print for non-value-literals")
        carg = arg[1]
        cargs.append(carg)
    return ["print_statement"] + cargs

def convert_statements(AST):
    cstatements = []
    statements = get_statements(AST)
    for statement in statements:
        tag, rest = statement[0], statement[1:]
        print tag
        try:
            if tag == "assignment_statement":
                cstatement = convert_assignment(rest)
                print cstatement
                cstatements.append(cstatement)
            print "HERE!?", repr(tag)
            if tag == "print_statement":
                print "Flooble"
                cstatement = convert_print(rest)
                cstatements.append(cstatement)

        except CannotConvert:
            pass
    return cstatements

def ast_to_cst(program_name, AST):
    cst = {}

    # Extract and handle variables
    pvariables = find_variables(AST)
    cvariables = []
    ctypes = {}
    includes = []
    for name in pvariables:
        ctype = python_type_to_c_type(pvariables[name])
        identifier = [ "identifier", ctype, name ]
        cvariables.append(identifier)
        ctypes[ctype] = True

    cstatements = convert_statements(AST)
    print cstatements

    # Based on variables, update includes
    for ctype in ctypes:
        inc = includes_for_ctype(ctype)
        if inc:
            includes.append(inc)

    # Based on statements, update includes
    for cstatement in cstatements:
        inc = includes_for_cstatement(cstatement)
        if inc:
            includes.append(inc)

    program = {}
    program["name"] = program_name
    program["includes"] = sorted(includes)
    program["main"] = {}
    program["main"]["c_frame"] = {}
    program["main"]["c_frame"]["identifiers"] = cvariables
    program["main"]["c_frame"]["statements"] = cstatements

    return { "PROGRAM" : program }


if __name__ == "__main__":
    AST =  ['program',
             ['statements',
              [['assignment_statement',
                ['greeting', 'IDENTIFIER', 1],
                ['ASSIGN', '='],
                ['value_literal', 'hello', 'STRING', 1]],
               ['assignment_statement',
                ['name', 'IDENTIFIER', 2],
                ['ASSIGN', '='],
                ['value_literal', 'world', 'STRING', 2]],
               ['print_statement',
                [['value_literal', 'greeting', 'IDENTIFIER', 4],
                 ['value_literal', 'name', 'IDENTIFIER', 4]]]]]]
    expect = {
                 "PROGRAM": {
                     "name": "hello",
                     "includes": [ "<iostream>", "<string>" ],
                     "main": {
                         "c_frame": {
                             "identifiers": [
                                 [ "identifier", "string", "greeting" ],
                                 [ "identifier", "string", "name" ]
                             ],
                             "statements": [
                                 [ "assignment", "greeting", "=", "\"hello\"" ],
                                 [ "assignment", "name", "=", "\"world\"" ],
                                 [ "print_statement", "greeting", "name" ]
                             ]
                         }
                     }
                 }
             }

    actual = ast_to_cst("hello", AST)

    print "actual == expect --->", actual == expect

    import json
    import pprint
    # print json.dumps(actual, indent=4)
    pprint.pprint(actual)




#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ply

# ------------------------------------------------------------
import ply.lex as lex


keywords = [ "True", "False","print" ]

tokens = [
   'NUMBER',
   'FLOAT',
   'HEX',
   'OCTAL',
   'BINARY',
   'STRING',
   'IDENTIFIER',
   'BOOLEAN',
   'EOL'
   ]
symbols = [ "ASSIGN", "COMMA" ]

tokens += [ x.upper() for x in keywords if (x not in [ "True", "False" ])]
tokens += symbols

## Regular expression rules for simple tokens
#t_PLUS    = r'\+'
#t_MINUS   = r'-'
#t_TIMES   = r'\*'
#t_DIVIDE  = r'/'
#t_LPAREN  = r'\('
#t_RPAREN  = r'\)'

t_ASSIGN = r'='
t_COMMA = r','

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z0-9_]*'

    if t.value in keywords:
        if t.value in ["True", "False"]:
            t.type = "BOOLEAN"
            t.value = True if t.value == "True" else False
        else:
            t.type = t.value.upper()
    return t

def t_SQUOTESTRING(t):
    r"'([^\\']|(\\.))*'"
    t.value = t.value[1:-1]
    t.value = t.value.replace('\\\'', '\'')
    t.type = "STRING"
    return t

def t_DQUOTESTRING(t):
    r'"([^\\"]|(\\.))*"'
    t.value = t.value[1:-1]
    t.value = t.value.replace('\\\"', '\"')
    t.type = "STRING"
    return t

# A regular expression rule with some action code
def t_BINARY(t):
    r'0b\d+'
    t.value = int(t.value,2)
    return t

def t_OCTAL(t):
    r'0o\d+'
    t.value = int(t.value,8)
    return t

def t_HEX(t):
    r'0x([abcdef]|\d)+'
    t.value = int(t.value,16)
    return t

def t_FLOAT(t):
    r'\d+\.\d+'
    t.value = float(t.value)
    return t

def t_NUMBER(t):
    r'\d+'
    t.value = int(t.value)
    return t


# Define a rule so we can track line numbers
def t_EOL(t):
    r'\n+'
    t.lexer.lineno += len(t.value)
    return t

# A string containing ignored characters (spaces and tabs)
t_ignore  = ' \t'

# Error handling rule
def t_error(t):
    print "Illegal character '%s'" % t.value[0]
    t.lexer.skip(1)

# Build the lexer
lexer = lex.lex()



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from distutils.core import setup
from distutils.version import LooseVersion
import os

def is_package(path):
    return (
        os.path.isdir(path) and
        os.path.isfile(os.path.join(path, '__init__.py'))
        )

def find_packages(path, base="" ):
    """ Find all packages in path """
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if os.path.isdir(dir):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            if is_package( dir ):
                packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages


packages = find_packages(".")
package_names = packages.keys()

setup(name = "pyxie",
      version = "0.0.2",
      description = "Little Python to C++ Compiler",
      url='http://www.sparkslabs.com/michael/',
      author='Michael Sparks (sparkslabs)',
      author_email="sparks.m@gmail.com",
      license='Apache Software License',

      scripts = [
                  'bin/pyxie',
                ],

      packages = package_names,
      package_dir = packages,
      package_data={},
      long_description = """
Little Python to C++ Compiler
"""
      )



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

blank_line = ""

source = []
def Print(*args):
    y = " ".join([str(x) for x in args])
    source.append(y)

def mkStatement(statement_spec):
    ss = statement_spec
    if ss[0] == "assignment":
        return Assigment( ss[1], ss[3], ss[2])

    if ss[0] == "print_statement":
        return PrintStatement(*statement_spec[1:])


class C_Program(object):
    def __init__(self):
        self.includes = []
        self.main_cframe = C_Frame()
        self.name = ""

    @classmethod
    def fromjson(klass, json):
        program = klass()
        prog_desc = json["PROGRAM"]
        program.name = prog_desc["name"]
        program.includes = list(prog_desc["includes"])
        main_spec = prog_desc["main"]["c_frame"]
        for identifier in main_spec["identifiers"]:
            print "IDENTIFIERS"
            program.main_cframe.identifiers.append(Identifier(identifier[1], identifier[2]))

        for statement in main_spec["statements"]:
            conc_statement = mkStatement(statement)
            program.main_cframe.statements.append(conc_statement)


        return program

    def generate(self):
        frame_lines = self.main_cframe.concrete()
        for include in self.includes:
            Print( "#include "+ include )

        Print()
        Print("using namespace std;")
        Print()
        Print("int main(int argc, char *argv[])")
        Print("{")
        for line in frame_lines:
            Print("    "+ line)
        Print("    "+ "return 0;")
        Print("}")

    def json(self):
        return { "PROGRAM": {"name": self.name,
                             "includes" : self.includes,
                             "main": self.main_cframe.json()}
               }

class C_Frame(object):
    def __init__(self):
        self.identifiers = []
        self.statements = []

    def json(self):
        for y in self.statements:
            print y
        return {"c_frame": {"identifiers" : [ x.json() for x in self.identifiers ],
                            "statements" : [y.json() for y in self.statements ] }
               }

    def concrete(self):
        block = []
        for identifier in self.identifiers:
            decl_code = identifier.decl_code()
            block.append(decl_code + ";")

        block.append(blank_line)
        for statement in self.statements:
            if statement:
                code = statement.code()
                block.append(code + ";")
        return block

class Identifier(object):
    def __init__(self, ctype, name):
        self.ctype = ctype
        self.name = name

    def json(self):
        return ["identifier", self.ctype, self.name ]

    def decl_code(self):
        return self.ctype + " " + self.name

class Assigment(object):
    def __init__(self, lvalue, rvalue, assigntype="="):
        self.lvalue = lvalue
        self.rvalue = rvalue
        self.assigntype = assigntype

    def json(self):
        return ["assignment", self.lvalue, self.assigntype, self.rvalue ]

    def code(self):
        return self.lvalue + " "+self.assigntype+" " + self.rvalue

class PrintStatement(object):
    def __init__(self, *args):
        self.args = args

    def json(self):
        return ["print_statement" ] + list(self.args[:])

    def code(self):
        return "cout << " + " << \" \" << ".join(self.args) + " << endl"

makefile_tmpl = """
all :
	g++ %(filename)s.c -o %(filename)s
"""

def build_program(json):
    json = C_Program.fromjson(json)


if __name__ == "__main__":

    # Build an example concrete syntax tree

    program = C_Program()

    program.name = ("hello")
    program.includes.append("<iostream>")
    program.includes.append("<string>")

    main = program.main_cframe
    main.identifiers.append(Identifier("string", "greeting"))
    main.identifiers.append(Identifier("string", "name"))
    main.statements.append(Assigment("greeting", '"hello"'))
    main.statements.append(Assigment("name", '"world"'))
    main.statements.append(PrintStatement("greeting", "name"))

    import pprint

    progj = program.json()

    pprint.pprint(progj)

    program_clone = C_Program.fromjson(progj)
    progj2 = program_clone.json()

    program = program_clone

    print "--------------------------------------------------------------"
    pprint.pprint(progj2)

    program.generate()

    import time
    import pprint
    import os

    pprint.pprint(source, width=200)
    now = int(time.time())
    dirname = str(now - 1427000000)
    print "BUILDING PROGRAM", dirname
    os.mkdir(dirname)
    f = open(os.path.join(dirname,program.name+".c"), "w")
    for line in source:
        f.write(line)
        f.write("\n")
    f.close()

    makefile = makefile_tmpl % {"filename": program.name }
    f = open(os.path.join(dirname,"Makefile"), "w")
    f.write(makefile)
    f.close()

    os.chdir(dirname)
    os.system("make")

